import os
import pandas as pd
import logging

def read_stock_data(stock_code, data_dir):
    try:
        directory_path = os.path.join(data_dir, 'stock_data')
        dataframes = []
        for file_name in os.listdir(directory_path):
            if file_name.startswith(stock_code) and file_name.endswith('_data.csv'):
                file_path = os.path.join(directory_path, file_name)
                logging.info(f"Reading data from {file_path}")
                df = pd.read_csv(file_path)
                
                # 필수 열이 존재하는지 확인
                required_columns = ['open', 'high', 'low', 'close', 'volume', 'date']
                missing_columns = [col for col in required_columns if col not in df.columns]
                if missing_columns:
                    logging.error(f"Missing columns: {missing_columns} in {file_path}")
                    continue  # 현재 파일을 건너뛰고 다음 파일로 진행
                
                # 데이터 형식이 올바른지 확인 (예: 'close' 열에 숫자가 아닌 값이 포함된 경우)
                try:
                    df['close'] = df['close'].apply(lambda x: float(x))
                except ValueError as ve:
                    logging.error(f"Error converting 'close' column to float in {file_path}: {ve}")
                    # 문제의 데이터를 로깅합니다.
                    logging.error(f"Problematic values in 'close' column: {df['close'].tolist()}")
                    continue  # 현재 파일을 건너뛰고 다음 파일로 진행

                dataframes.append(df)
        if dataframes:
            combined_df = pd.concat(dataframes, ignore_index=True)
            # 데이터 출력
            logging.info(f"Combined DataFrame head: \n{combined_df.head()}")
            return combined_df
        else:
            logging.error(f"No valid data files found for stock {stock_code} in directory: {directory_path}")
            return pd.DataFrame()
    except pd.errors.EmptyDataError as e:
        logging.error(f"Empty data error for {stock_code}: {e}")
        return pd.DataFrame()
    except pd.errors.ParserError as e:
        logging.error(f"Parsing error for {stock_code}: {e}")
        return pd.DataFrame()
    except Exception as e:
        logging.error(f"Error reading stock data for {stock_code}: {e}")
        return None
