import os
import sys
import logging
import pandas as pd
from dotenv import load_dotenv
import time
from datetime import datetime

# 프로젝트 경로 설정
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
scripts_dir = os.path.join(project_root, 'scripts')
data_collection_dir = os.path.join(scripts_dir, 'data_collection')

# 필요한 디렉토리를 sys.path에 추가
sys.path.extend([
    scripts_dir,
    data_collection_dir,
    os.path.join(scripts_dir, 'data_processing'),
    os.path.join(scripts_dir, 'model'),
    os.path.join(scripts_dir, 'trading'),
    os.path.join(scripts_dir, 'utils')
])

# 환경 변수 로드 및 설정
CONFIG_DIR = 'C:\\Users\\user\\projects\\st\\config'
load_dotenv(dotenv_path=os.path.join(CONFIG_DIR, '.env'))

# 필요한 모듈 가져오기
try:
    from data_collection.stock_data_collector import fetch_and_save_data, get_stock_codes_and_names, get_valid_token
    from data_collection.news_data_collector import fetch_news_for_all_stocks, fetch_and_save_financial_data
    from data_processing.data_preprocessing import preprocess_data, preprocess_and_save
    from data_processing.feature_engineering import add_features
    from model.main_model_training import ModelTrainer
    from trading.backtesting import Backtester
    from utils.performance_monitoring import PerformanceMonitor
    from utils.utils import get_last_run_date, save_last_run_date, save_data  # 필요한 유틸리티 함수 추가
except ModuleNotFoundError as e:
    logging.error(f"Module not found: {e}")
    sys.exit(1)

# 로깅 설정
logging.basicConfig(level=logging.INFO)

def load_csv_files_from_directory(directory_path, identifier_col=None):
    dataframes = []
    for file_name in os.listdir(directory_path):
        if file_name.endswith('.csv'):
            file_path = os.path.join(directory_path, file_name)
            try:
                df = pd.read_csv(file_path, encoding='utf-8')
                logging.info(f"Loaded {file_name} with columns: {df.columns}")
                df.columns = df.columns.str.lower()  # Make all column names lowercase

                if identifier_col and identifier_col not in df.columns:
                    ticker = os.path.splitext(file_name)[0].split('_')[0]
                    df[identifier_col] = ticker

                if not df.empty:
                    logging.info(f"First few rows of {file_name}:\n{df.head()}")
                    dataframes.append(df)
                else:
                    logging.warning(f"Empty file: {file_path}")
            except pd.errors.EmptyDataError:
                logging.warning(f"Empty file: {file_path}")
            except Exception as e:
                logging.error(f"Error reading {file_path}: {e}")
    if dataframes:
        return pd.concat(dataframes, ignore_index=True)
    else:
        logging.error(f"No valid CSV files found in directory: {directory_path}")
        return pd.DataFrame()

def main():
    logging.info("Starting the main data collection process...")

    # Step 1: 토큰 갱신 및 유효성 검사
    logging.info("Step 1: Checking token validity...")
    token = get_valid_token()
    if not token:
        logging.error("Failed to retrieve a valid token. Exiting.")
        return

    # Step 2: 조건 검색을 통해 종목 코드 불러오기
    logging.info("Step 2: Fetching stock codes using search conditions...")
    stock_codes_and_names = get_stock_codes_and_names(seq=0)  # 필요한 경우 seq 값을 변경하며 호출
    
    if not stock_codes_and_names:
        logging.error("Failed to retrieve any stock codes. Exiting.")
        return

    # Step 3: 불러온 종목 코드에 대해 데이터 수집
    logging.info("Step 3: Fetching stock data for detected stocks...")
    fetch_and_save_data(stock_codes_and_names)
    
    # Step 4: 뉴스와 재무 데이터 수집
    logging.info("Step 4: Fetching news and financial data...")
    fetch_news_for_all_stocks()
    fetch_and_save_financial_data()

    # Step 5: 데이터 전처리 및 저장
    logging.info("Step 5: Preprocessing the data and saving it...")
    data_dir = os.path.join(project_root, 'data')
    stock_data = load_csv_files_from_directory(os.path.join(data_dir, 'stock_data'), identifier_col='corp_code')
    if stock_data.empty:
        logging.error("Stock data is missing or empty after loading.")
        return

    financial_data = load_csv_files_from_directory(os.path.join(data_dir, 'financial_data'))
    economic_data = load_csv_files_from_directory(os.path.join(data_dir, 'economic_data'))

    if financial_data.empty:
        logging.warning("Financial data is missing or empty.")
    if economic_data.empty:
        logging.warning("Economic data is missing or empty.")

    # 데이터 전처리 및 저장
    output_dir = os.path.join(data_dir, 'preprocessed_data')
    preprocess_and_save(stock_data, financial_data, economic_data, output_dir, 'your_corp_code')

    # 데이터 전처리 후 피처와 타겟을 나누고 스케일링
    logging.info("Preprocessing the data for model training...")
    result = preprocess_data(stock_data, financial_data, economic_data)
    if result is None:
        logging.error("Preprocessing failed. Exiting.")
        return
    
    (X_train, X_test, y_train, y_test), scaler, processed_data = result

    # Step 6: Model Training
    logging.info("Step 6: Starting model training...")
    model_trainer = ModelTrainer()
    trained_model = model_trainer.train(X_train, y_train, X_test, y_test)
    
    if trained_model is None:
        logging.error("Model training failed. Exiting.")
        return

    # Step 7: Backtesting
    logging.info("Step 7: Starting backtesting...")
    backtester = Backtester()
    backtest_results = backtester.run_backtest(trained_model, X_test, y_test)

    if backtest_results is None:
        logging.error("Backtesting failed. Exiting.")
        return

    # Step 8: Performance Monitoring
    logging.info("Step 8: Monitoring performance...")
    performance_monitor = PerformanceMonitor()
    performance_monitor.monitor(backtest_results)

    logging.info("Main workflow completed successfully.")

if __name__ == "__main__":
    main()
